# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import gc
import importlib.util
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")

TEST_DATA_PATH = Path(__file__).parent / "data" / "trivista"

testfile_spec_path = TEST_DATA_PATH / "spec_1s_1acc_1frame_average.tvf"
testfile_spec_2frames_path = TEST_DATA_PATH / "spec_3s_1acc_2frames_average.tvf"
testfile_spec_2acc_path = TEST_DATA_PATH / "spec_3s_2acc_1frame_average.tvf"
testfile_spec_2acc_no_average_path = (
    TEST_DATA_PATH / "spec_3s_2acc_1frame_no_average.tvf"
)
testfile_spec_timeseries_path = TEST_DATA_PATH / "spec_timeseries_2x1s_delta3s.tvf"
testfile_map_path = TEST_DATA_PATH / "map.tvf"
testfile_step_and_glue_path = TEST_DATA_PATH / "spec_step_and_glue.tvf"
testfile_triple_add_path = TEST_DATA_PATH / "spec_multiple_spectrometers.tvf"
testfile_linescan_path = TEST_DATA_PATH / "linescan.tvf"


if importlib.util.find_spec("lumispy") is None:
    lumispy_installed = False
else:
    lumispy_installed = True


class TestSpec:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_spec_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
        )
        cls.s_non_uniform_unfiltered = hs.load(
            testfile_spec_path,
            reader="TriVista",
            use_uniform_signal_axis=False,
            filter_original_metadata=False,
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform_unfiltered
        gc.collect()

    def test_data(self):
        expected_data = np.array([27732, 39135, 35343, -420, 32464])
        np.testing.assert_allclose(expected_data, self.s.isig[:5].data)
        np.testing.assert_allclose(
            expected_data, self.s_non_uniform_unfiltered.isig[:5].data
        )

    def test_axes(self):
        expected_axis = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 1024,
            }
        }

        expected_axis_non_uniform = {
            "axis-0": {
                "_type": "DataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
            }
        }

        expected_values_non_uniform_axis = np.array(
            [794.220731002166, 794.261531331698, 794.302331149879]
        )

        uniform_axes_manager = deepcopy(self.s.axes_manager.as_dictionary())
        non_uniform_axis_manager = deepcopy(
            self.s_non_uniform_unfiltered.axes_manager.as_dictionary()
        )

        np.testing.assert_allclose(
            uniform_axes_manager["axis-0"].pop("scale"), 0.040515, atol=0.000005
        )
        np.testing.assert_allclose(
            uniform_axes_manager["axis-0"].pop("offset"), 794.22, atol=0.1
        )
        np.testing.assert_allclose(
            expected_values_non_uniform_axis,
            non_uniform_axis_manager["axis-0"].pop("axis")[:3],
        )
        assert expected_axis == uniform_axes_manager
        assert expected_axis_non_uniform == non_uniform_axis_manager

    def test_metadata(self):
        metadata = deepcopy(self.s.metadata.as_dictionary())
        metadata_non_uniform = deepcopy(
            self.s_non_uniform_unfiltered.metadata.as_dictionary()
        )

        assert metadata["General"]["FileIO"]["0"]["io_plugin"] == "rsciio.trivista"
        assert (
            metadata_non_uniform["General"]["FileIO"]["0"]["io_plugin"]
            == "rsciio.trivista"
        )

        assert metadata["General"]["date"] == "2022-06-14"
        assert (
            metadata["General"]["original_filename"]
            == "spec_1s_1acc_1frame_average.tvf"
        )
        assert metadata["General"]["time"] == "13:34:27"
        assert (
            metadata["General"]["title"]
            == metadata["General"]["original_filename"][:-4]
        )

        assert metadata["Signal"]["quantity"] == "Intensity (Counts)"
        if lumispy_installed:
            assert metadata["Signal"]["signal_type"] == "Luminescence"
        else:
            assert metadata["Signal"]["signal_type"] == ""

        assert metadata["Acquisition_instrument"]["Detector"]["glued_spectrum"] is False
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"]["calc_average"]
            == "True"
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["exposure_per_frame"], 1
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["frames"], 1
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["integration_time"], 1
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["temperature"], -25
        )

        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Laser"]["objective_magnification"], 100
        )

        assert (
            metadata["Acquisition_instrument"]["Spectrometer"]["Grating"][
                "blazing_wavelength"
            ]
            == "H-NIR"
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["Grating"][
                "groove_density"
            ],
            750,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["central_wavelength"],
            815,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["entrance_slit_width"],
            0.2,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["exit_slit_width"], 0
        )
        assert (
            metadata["Acquisition_instrument"]["Spectrometer"]["model"] == "SP-2-750i"
        )

        del metadata["General"]["FileIO"]
        del metadata_non_uniform["General"]["FileIO"]

        assert metadata == metadata_non_uniform

    def test_original_metadata(self):
        expected_metadata = {
            "XmlMain": {
                "Version": "1.0.1",
                "Filename": "C:\\Users\\optik\\Desktop\\Testspectrum.tvf",
                "DateTime": "06/14/2022 13:35:01",
                "PlugInName": "",
                "PlugInData": "",
                "FileInfoSerialized": {"Info": {"Groups": None}},
            },
            "Document": {
                "Version": "2",
                "Label": "Intensity",
                "DataLabel": "Counts",
                "DocType": "Spectra",
                "RecordTime": "06/14/2022 13:34:27.453",
                "ModeName": "",
                "DataSource": "nothing",
                "Encoding": "Ascii",
                "ColorMode": "Grayscale",
                "ViewDisplayMode": "Graph",
                "ViewImageColorMode": "False",
                "InfoSerialized": {
                    "Document": {"Record Time": "6/14/2022 1:34 PM"},
                    "Experiment": {"Used Setup": "PL_Stage3_750g"},
                    "Spectrometers": {
                        "Spectrometer": {
                            "Serialnumber": "27580185",
                            "Model": "SP-2-750i",
                            "Stage_Number": "1",
                            "Focallength": "749",
                            "Inclusion_Angle": "6.5",
                            "Detector_Angle": "0.68",
                            "Groove_Density": "750 g/mm",
                            "Order": "1",
                            "Slit_Entrance-Front": "200",
                            "Slit_Entrance-Side": "0",
                            "Slit_Exit-Front": "0",
                            "Slit_Exit-Side": "0",
                        }
                    },
                    "Detector": {
                        "Name": "Camera1",
                        "Serialnumber": None,
                        "Detector_Size": "1024;1",
                        "Detector_Temperature": "-25",
                        "Exposure_Time_(ms)": "1000",
                        "Exposure_Mode": None,
                        "No_of_Accumulations": "1",
                        "Calc_Average": "True",
                        "No_of_Frames": "1",
                        "ADC__Readout_Port": "Normal",
                        "ADC__Rate_Resolution": "1 MHz",
                        "ADC__Gain": "2",
                        "Clearing__Mode": None,
                        "Clearing__No_of_Cleans": "1",
                        "Region_of_Interests": "1|1;1024;1;1;1;1",
                    },
                    "Calibration": {
                        "Center_Wavelength": "815.000",
                        "Laser_Wavelength": "0.000",
                    },
                },
            },
            "Hardware": {
                "Spectrometers": {
                    "Spectrometer": {
                        "Gratings": {
                            "Grating": {
                                "Offsets": {"Count": "0"},
                                "GrooveDensity": "750 g/mm",
                                "Blaze": "H-NIR",
                                "Turret": "1",
                            },
                            "Count": "3",
                        },
                        "MirrorEntrance": {
                            "Name": "M1",
                            "State": "isMotorized",
                            "Position": "front",
                        },
                        "MirrorExit": {
                            "Name": "M2",
                            "State": "isMotorized",
                            "Position": "side",
                        },
                        "SlitEntranceSide": {
                            "Name": "S1",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitEntranceFront": {
                            "Name": "S2",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "500",
                        },
                        "SlitExitFront": {
                            "Name": "S3",
                            "State": "disabled",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitExitSide": {
                            "Name": "S4",
                            "State": "disabled",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "FilterWheel": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "0",
                            "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                        },
                        "Shutter": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "open",
                            "ShutterMode": "OpenForExperiment",
                            "ShutterClosedForFilterwheel": "True",
                            "ShutterClosedForSetupChange": "True",
                            "ShutterClosedForLaserCrossing": "True",
                            "ShutterClosedForLaserCrossingValue": "0.000",
                            "ShutterClosedForLaserCrossingUnit": "Nanometer",
                        },
                        "GUID": "37a2855b-db41-4268-970b-bb75d96e5654",
                        "Model": "SP-2-750i",
                        "Serialnumber": "27580185",
                        "FocalLength": "749",
                        "InclusionAngle": "6.500",
                        "InclusionAngleAdditive": "39.000",
                        "DetectorAngle": "0.680",
                        "PixelWidthExitFront": "25.000",
                        "PixelWidthExitSide": "-1.000",
                        "Backlash": "ScanUp",
                        "DriveDirection": "Normal",
                        "ComPortSpec": "COM5",
                        "ComPortEntrance": "",
                        "ComPortExit": "",
                        "DemoMode": "False",
                        "UsedAsLightsource": "False",
                        "UseBackwards": "False",
                        "Wavelength": "850.184",
                        "Grating": "2",
                        "ReadTimeout": "35000",
                        "Activated": "False",
                        "IsFixedCalibration": "False",
                        "FixedCalibrationPoints": "0",
                    },
                    "Count": "3",
                },
                "Detectors": {
                    "Detector": {
                        "GUID": "c1bc9fd2-135e-4098-b681-d30f6ea8be82",
                        "Driver": "PvCam32 Driver",
                        "Name": "Camera1",
                        "DisplayName": "Camera1",
                        "SerialNumber": "",
                        "DetectorType": "None",
                        "Size": "1024;1",
                        "OpenAtStartup": "False",
                        "Usage": "Default",
                    },
                    "Count": "3",
                },
                "SinglePointDetectors": {"Count": "0"},
                "LightSources": {
                    "LightSource": [
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "0.000"},
                            "GUID": "a57ac71a-c667-4501-a24a-596bc54b2c8c",
                            "Name": "Calibration Lamp",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "0"},
                            "GUID": "cac03351-735e-4592-9bb2-c17f64b3649f",
                            "Name": "Calibration Lamp2",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "0"},
                            "GUID": "84a08bee-227a-46f0-b1e1-0f171808a155",
                            "Name": "Calibration Lamp3",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "0.000"},
                            "GUID": "a0601d22-23a0-4587-bb79-35d5823916c4",
                            "Name": "Laser",
                            "Serialnumber": "",
                            "FixedWavelength": "False",
                            "CanChange": "False",
                            "Type": "Laser",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "790.000"},
                            "GUID": "a36b29da-0d54-4ab7-860c-967269f76217",
                            "Name": "tisa",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "True",
                            "Type": "Laser",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                    ],
                    "Count": "5",
                },
                "ArduinoHardware": {"Count": "0"},
                "Additionals": {"Count": "0"},
                "MotorizedStages": {
                    "MotorizedStage": {
                        "GUID": "c551372e-56af-4686-b7c1-48ca00dfdf27",
                        "DriverName": "Tango Driver",
                        "Name": "XY-Stage",
                        "ConnectionString": "COM15",
                        "Serialnumber": "144012056",
                        "IsTriggerModeAvailable": "False",
                    },
                    "Count": "1",
                },
                "Microscopes": {
                    "Microscope": {
                        "Objectives": {
                            "Objective": {
                                "IsEnabled": "True",
                                "Name": "100x",
                                "Magnification": "100.000",
                                "FactorMeasuredWithResolution": "1280",
                                "Factor": "34.700",
                                "LaserOffsetX": "0.000",
                                "LaserOffsetY": "0.000",
                                "OffsetX": "0.000",
                                "OffsetY": "0.000",
                                "LineLaserOffsetX": "0.000",
                                "LineLaserOffsetY": "0.000",
                                "LineOffsetX": "0.000",
                                "LineOffsetY": "0.000",
                                "LineFocusFactor": "1.130",
                            },
                            "Count": "10",
                        },
                        "GUID": "841bfef9-b969-43f0-9ca3-2522c223a31f",
                        "Name": "Microscope",
                        "StageID": "c551372e-56af-4686-b7c1-48ca00dfdf27",
                        "ImagingCameraID": "",
                        "RamanBoxAvailable": "True",
                        "RamanBoxLinearStage.Available": "False",
                        "RamanBoxLinearStage.ConnectedHarwareID": "",
                        "RamanBoxFilterWheel.Available": "False",
                        "RamanBoxFilterWheel.ConnectedHarwareID": "",
                        "RamanBoxBeamSplitter.Available": "False",
                        "RamanBoxBeamSplitter.ConnectedHarwareID": "",
                        "RamanBoxNotchFilter.Available": "False",
                        "RamanBoxNotchFilter.ConnectedHarwareID": "",
                        "ActiveObjectiveIndex": "0",
                    },
                    "Count": "2",
                },
                "Cryostats": {"Count": "0"},
                "Version": "2",
            },
        }

        assert expected_metadata == self.s.original_metadata.as_dictionary()

    def test_unfiltered_original_metadata(self):
        expected_metadata = {
            "XmlMain": {
                "Version": "1.0.1",
                "Filename": "C:\\Users\\optik\\Desktop\\Testspectrum.tvf",
                "DateTime": "06/14/2022 13:35:01",
                "PlugInName": "",
                "PlugInData": "",
                "FileInfoSerialized": {"Info": {"Groups": None}},
            },
            "Document": {
                "Version": "2",
                "Label": "Intensity",
                "DataLabel": "Counts",
                "DocType": "Spectra",
                "RecordTime": "06/14/2022 13:34:27.453",
                "ModeName": "",
                "DataSource": "nothing",
                "Encoding": "Ascii",
                "ColorMode": "Grayscale",
                "ViewDisplayMode": "Graph",
                "ViewImageColorMode": "False",
                "InfoSerialized": {
                    "Info": {
                        "Groups": {
                            "Group": [
                                {
                                    "Name": "Document",
                                    "Items": {
                                        "Item": {
                                            "Name": "Record Time",
                                            "Value": "6/14/2022 1:34 PM",
                                            "IsVisible": "true",
                                        }
                                    },
                                    "Groups": None,
                                    "IsVisible": "true",
                                },
                                {
                                    "Name": "Experiment",
                                    "Items": {
                                        "Item": {
                                            "Name": "Used Setup",
                                            "Value": "PL_Stage3_750g",
                                            "IsVisible": "true",
                                        }
                                    },
                                    "Groups": None,
                                    "IsVisible": "true",
                                },
                                {
                                    "Name": "Spectrometers",
                                    "Items": None,
                                    "Groups": {
                                        "Group": {
                                            "Name": "Spectrometer",
                                            "Items": {
                                                "Item": [
                                                    {
                                                        "Name": "Serialnumber",
                                                        "Value": "27580185",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Model",
                                                        "Value": "SP-2-750i",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Stage_Number",
                                                        "Value": "1",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Focallength",
                                                        "Value": "749",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Inclusion_Angle",
                                                        "Value": "6.5",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Detector_Angle",
                                                        "Value": "0.68",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Groove_Density",
                                                        "Value": "750 g/mm",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Order",
                                                        "Value": "1",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Slit_Entrance-Front",
                                                        "Value": "200",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Slit_Entrance-Side",
                                                        "Value": "0",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Slit_Exit-Front",
                                                        "Value": "0",
                                                        "IsVisible": "true",
                                                    },
                                                    {
                                                        "Name": "Slit_Exit-Side",
                                                        "Value": "0",
                                                        "IsVisible": "true",
                                                    },
                                                ]
                                            },
                                            "Groups": None,
                                            "IsVisible": "true",
                                        }
                                    },
                                    "IsVisible": "true",
                                },
                                {
                                    "Name": "Detector",
                                    "Items": {
                                        "Item": [
                                            {
                                                "Name": "Name",
                                                "Value": "Camera1",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Serialnumber",
                                                "Value": None,
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Detector_Size",
                                                "Value": "1024;1",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Detector_Temperature",
                                                "Value": "-25",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Exposure_Time_(ms)",
                                                "Value": "1000",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Exposure_Mode",
                                                "Value": None,
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "No_of_Accumulations",
                                                "Value": "1",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Calc_Average",
                                                "Value": "True",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "No_of_Frames",
                                                "Value": "1",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "ADC__Readout_Port",
                                                "Value": "Normal",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "ADC__Rate_Resolution",
                                                "Value": "1 MHz",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "ADC__Gain",
                                                "Value": "2",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Clearing__Mode",
                                                "Value": None,
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Clearing__No_of_Cleans",
                                                "Value": "1",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Region_of_Interests",
                                                "Value": "1|1;1024;1;1;1;1",
                                                "IsVisible": "true",
                                            },
                                        ]
                                    },
                                    "Groups": None,
                                    "IsVisible": "true",
                                },
                                {
                                    "Name": "Calibration",
                                    "Items": {
                                        "Item": [
                                            {
                                                "Name": "Center_Wavelength",
                                                "Value": "815.000",
                                                "IsVisible": "true",
                                            },
                                            {
                                                "Name": "Laser_Wavelength",
                                                "Value": "0.000",
                                                "IsVisible": "true",
                                            },
                                        ]
                                    },
                                    "Groups": None,
                                    "IsVisible": "true",
                                },
                            ]
                        }
                    }
                },
            },
            "Hardware": {
                "Spectrometers": {
                    "Spectrometer": [
                        {
                            "Gratings": {
                                "Grating": [
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "1800 g/mm",
                                        "Blaze": "H-VIS",
                                        "Turret": "1",
                                    },
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "900 g/mm",
                                        "Blaze": "550NM",
                                        "Turret": "1",
                                    },
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "750 g/mm",
                                        "Blaze": "H-NIR",
                                        "Turret": "1",
                                    },
                                ],
                                "Count": "3",
                            },
                            "MirrorEntrance": {
                                "Name": "M1",
                                "State": "isMotorized",
                                "Position": "front",
                            },
                            "MirrorExit": {
                                "Name": "M2",
                                "State": "isMotorized",
                                "Position": "side",
                            },
                            "SlitEntranceSide": {
                                "Name": "S1",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "SlitEntranceFront": {
                                "Name": "S2",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "200",
                            },
                            "SlitExitFront": {
                                "Name": "S3",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "SlitExitSide": {
                                "Name": "S4",
                                "State": "disabled",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "FilterWheel": {
                                "Name": "",
                                "State": "disabled",
                                "Position": "0",
                                "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                            },
                            "Shutter": {
                                "Name": "",
                                "State": "disabled",
                                "Position": "open",
                                "ShutterMode": "OpenForExperiment",
                                "ShutterClosedForFilterwheel": "True",
                                "ShutterClosedForSetupChange": "True",
                                "ShutterClosedForLaserCrossing": "True",
                                "ShutterClosedForLaserCrossingValue": "0.000",
                                "ShutterClosedForLaserCrossingUnit": "Nanometer",
                            },
                            "GUID": "c05fbec6-7cc0-4546-a009-06850868f43d",
                            "Model": "SP-2-500i",
                            "Serialnumber": "25580419",
                            "FocalLength": "500",
                            "InclusionAngle": "8.600",
                            "InclusionAngleAdditive": "51.600",
                            "DetectorAngle": "0.000",
                            "PixelWidthExitFront": "-1.000",
                            "PixelWidthExitSide": "-1.000",
                            "Backlash": "ScanUp",
                            "DriveDirection": "Normal",
                            "ComPortSpec": "COM3",
                            "ComPortEntrance": "",
                            "ComPortExit": "",
                            "DemoMode": "False",
                            "UsedAsLightsource": "False",
                            "UseBackwards": "False",
                            "Wavelength": "925.010",
                            "Grating": "2",
                            "ReadTimeout": "35000",
                            "Activated": "False",
                            "IsFixedCalibration": "False",
                            "FixedCalibrationPoints": "0",
                        },
                        {
                            "Gratings": {
                                "Grating": [
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "1800 g/mm",
                                        "Blaze": "H-VIS",
                                        "Turret": "1",
                                    },
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "900 g/mm",
                                        "Blaze": "550NM",
                                        "Turret": "1",
                                    },
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "750 g/mm",
                                        "Blaze": "H-NIR",
                                        "Turret": "1",
                                    },
                                ],
                                "Count": "3",
                            },
                            "MirrorEntrance": {
                                "Name": "M1",
                                "State": "isMotorized",
                                "Position": "side",
                            },
                            "MirrorExit": {
                                "Name": "M2",
                                "State": "isMotorized",
                                "Position": "side",
                            },
                            "SlitEntranceSide": {
                                "Name": "S1",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "10000",
                            },
                            "SlitEntranceFront": {
                                "Name": "S2",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "SlitExitFront": {
                                "Name": "S3",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "SlitExitSide": {
                                "Name": "S4",
                                "State": "disabled",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "FilterWheel": {
                                "Name": "",
                                "State": "disabled",
                                "Position": "0",
                                "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                            },
                            "Shutter": {
                                "Name": "",
                                "State": "disabled",
                                "Position": "open",
                                "ShutterMode": "OpenForExperiment",
                                "ShutterClosedForFilterwheel": "True",
                                "ShutterClosedForSetupChange": "True",
                                "ShutterClosedForLaserCrossing": "True",
                                "ShutterClosedForLaserCrossingValue": "0.000",
                                "ShutterClosedForLaserCrossingUnit": "Nanometer",
                            },
                            "GUID": "39ade9ad-caa8-48c2-a3d0-4f79cc44bda7",
                            "Model": "SP-2-500i",
                            "Serialnumber": "25580420",
                            "FocalLength": "500",
                            "InclusionAngle": "8.600",
                            "InclusionAngleAdditive": "51.600",
                            "DetectorAngle": "0.000",
                            "PixelWidthExitFront": "1.000",
                            "PixelWidthExitSide": "-1.000",
                            "Backlash": "ScanUp",
                            "DriveDirection": "Normal",
                            "ComPortSpec": "COM4",
                            "ComPortEntrance": "",
                            "ComPortExit": "",
                            "DemoMode": "False",
                            "UsedAsLightsource": "False",
                            "UseBackwards": "False",
                            "Wavelength": "-924.910",
                            "Grating": "2",
                            "ReadTimeout": "35000",
                            "Activated": "False",
                            "IsFixedCalibration": "False",
                            "FixedCalibrationPoints": "0",
                        },
                        {
                            "Gratings": {
                                "Grating": [
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "1800 g/mm",
                                        "Blaze": "H-VIS",
                                        "Turret": "1",
                                    },
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "900 g/mm",
                                        "Blaze": "550NM",
                                        "Turret": "1",
                                    },
                                    {
                                        "Offsets": {"Count": "0"},
                                        "GrooveDensity": "750 g/mm",
                                        "Blaze": "H-NIR",
                                        "Turret": "1",
                                    },
                                ],
                                "Count": "3",
                            },
                            "MirrorEntrance": {
                                "Name": "M1",
                                "State": "isMotorized",
                                "Position": "front",
                            },
                            "MirrorExit": {
                                "Name": "M2",
                                "State": "isMotorized",
                                "Position": "side",
                            },
                            "SlitEntranceSide": {
                                "Name": "S1",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "SlitEntranceFront": {
                                "Name": "S2",
                                "State": "isMotorized",
                                "MaxWidth": "12000",
                                "Position": "500",
                            },
                            "SlitExitFront": {
                                "Name": "S3",
                                "State": "disabled",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "SlitExitSide": {
                                "Name": "S4",
                                "State": "disabled",
                                "MaxWidth": "12000",
                                "Position": "0",
                            },
                            "FilterWheel": {
                                "Name": "",
                                "State": "disabled",
                                "Position": "0",
                                "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                            },
                            "Shutter": {
                                "Name": "",
                                "State": "disabled",
                                "Position": "open",
                                "ShutterMode": "OpenForExperiment",
                                "ShutterClosedForFilterwheel": "True",
                                "ShutterClosedForSetupChange": "True",
                                "ShutterClosedForLaserCrossing": "True",
                                "ShutterClosedForLaserCrossingValue": "0.000",
                                "ShutterClosedForLaserCrossingUnit": "Nanometer",
                            },
                            "GUID": "37a2855b-db41-4268-970b-bb75d96e5654",
                            "Model": "SP-2-750i",
                            "Serialnumber": "27580185",
                            "FocalLength": "749",
                            "InclusionAngle": "6.500",
                            "InclusionAngleAdditive": "39.000",
                            "DetectorAngle": "0.680",
                            "PixelWidthExitFront": "25.000",
                            "PixelWidthExitSide": "-1.000",
                            "Backlash": "ScanUp",
                            "DriveDirection": "Normal",
                            "ComPortSpec": "COM5",
                            "ComPortEntrance": "",
                            "ComPortExit": "",
                            "DemoMode": "False",
                            "UsedAsLightsource": "False",
                            "UseBackwards": "False",
                            "Wavelength": "850.184",
                            "Grating": "2",
                            "ReadTimeout": "35000",
                            "Activated": "False",
                            "IsFixedCalibration": "False",
                            "FixedCalibrationPoints": "0",
                        },
                    ],
                    "Count": "3",
                },
                "Detectors": {
                    "Detector": [
                        {
                            "GUID": "c1bc9fd2-135e-4098-b681-d30f6ea8be82",
                            "Driver": "PvCam32 Driver",
                            "Name": "Camera1",
                            "DisplayName": "Camera1",
                            "SerialNumber": "",
                            "DetectorType": "None",
                            "Size": "1024;1",
                            "OpenAtStartup": "False",
                            "Usage": "Default",
                        },
                        {
                            "GUID": "bbfbd5b4-6ea8-43de-8760-3b91c5867333",
                            "Driver": "SpectraHub-Driver",
                            "Name": "Spectra-Hub_COM7",
                            "DisplayName": "Name",
                            "SerialNumber": "210367",
                            "DetectorType": "SinglePointDetector",
                            "Size": "1;1",
                            "OpenAtStartup": "False",
                            "Usage": "Default",
                        },
                        {
                            "GUID": "98abd847-6843-4c4c-b88c-7340de27ce13",
                            "Driver": "SpectraHub-Driver",
                            "Name": "Spectra-Hub_COM8",
                            "DisplayName": "Name",
                            "SerialNumber": "210238",
                            "DetectorType": "SinglePointDetector",
                            "Size": "1;1",
                            "OpenAtStartup": "False",
                            "Usage": "Default",
                        },
                    ],
                    "Count": "3",
                },
                "SinglePointDetectors": {"Count": "0"},
                "LightSources": {
                    "LightSource": [
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "0.000"},
                            "GUID": "a57ac71a-c667-4501-a24a-596bc54b2c8c",
                            "Name": "Calibration Lamp",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "0"},
                            "GUID": "cac03351-735e-4592-9bb2-c17f64b3649f",
                            "Name": "Calibration Lamp2",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "0"},
                            "GUID": "84a08bee-227a-46f0-b1e1-0f171808a155",
                            "Name": "Calibration Lamp3",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "0.000"},
                            "GUID": "a0601d22-23a0-4587-bb79-35d5823916c4",
                            "Name": "Laser",
                            "Serialnumber": "",
                            "FixedWavelength": "False",
                            "CanChange": "False",
                            "Type": "Laser",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "790.000"},
                            "GUID": "a36b29da-0d54-4ab7-860c-967269f76217",
                            "Name": "tisa",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "True",
                            "Type": "Laser",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                    ],
                    "Count": "5",
                },
                "ArduinoHardware": {"Count": "0"},
                "Additionals": {"Count": "0"},
                "MotorizedStages": {
                    "MotorizedStage": {
                        "GUID": "c551372e-56af-4686-b7c1-48ca00dfdf27",
                        "DriverName": "Tango Driver",
                        "Name": "XY-Stage",
                        "ConnectionString": "COM15",
                        "Serialnumber": "144012056",
                        "IsTriggerModeAvailable": "False",
                    },
                    "Count": "1",
                },
                "Microscopes": {
                    "Microscope": [
                        {
                            "Objectives": {
                                "Objective": [
                                    {
                                        "IsEnabled": "True",
                                        "Name": "100x",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                ],
                                "Count": "10",
                            },
                            "GUID": "841bfef9-b969-43f0-9ca3-2522c223a31f",
                            "Name": "Microscope",
                            "StageID": "c551372e-56af-4686-b7c1-48ca00dfdf27",
                            "ImagingCameraID": "",
                            "RamanBoxAvailable": "True",
                            "RamanBoxLinearStage.Available": "False",
                            "RamanBoxLinearStage.ConnectedHarwareID": "",
                            "RamanBoxFilterWheel.Available": "False",
                            "RamanBoxFilterWheel.ConnectedHarwareID": "",
                            "RamanBoxBeamSplitter.Available": "False",
                            "RamanBoxBeamSplitter.ConnectedHarwareID": "",
                            "RamanBoxNotchFilter.Available": "False",
                            "RamanBoxNotchFilter.ConnectedHarwareID": "",
                            "ActiveObjectiveIndex": "0",
                        },
                        {
                            "Objectives": {
                                "Objective": [
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                    {
                                        "IsEnabled": "False",
                                        "Name": "",
                                        "Magnification": "100.000",
                                        "FactorMeasuredWithResolution": "1280",
                                        "Factor": "34.700",
                                        "LaserOffsetX": "0.000",
                                        "LaserOffsetY": "0.000",
                                        "OffsetX": "0.000",
                                        "OffsetY": "0.000",
                                        "LineLaserOffsetX": "0.000",
                                        "LineLaserOffsetY": "0.000",
                                        "LineOffsetX": "0.000",
                                        "LineOffsetY": "0.000",
                                        "LineFocusFactor": "1.130",
                                    },
                                ],
                                "Count": "10",
                            },
                            "GUID": "ee496960-089d-4040-a5ca-cde5f9d1b0f8",
                            "Name": "Microscope",
                            "StageID": "",
                            "ImagingCameraID": "",
                            "RamanBoxAvailable": "False",
                            "RamanBoxLinearStage.Available": "False",
                            "RamanBoxLinearStage.ConnectedHarwareID": "",
                            "RamanBoxFilterWheel.Available": "False",
                            "RamanBoxFilterWheel.ConnectedHarwareID": "",
                            "RamanBoxBeamSplitter.Available": "False",
                            "RamanBoxBeamSplitter.ConnectedHarwareID": "",
                            "RamanBoxNotchFilter.Available": "False",
                            "RamanBoxNotchFilter.ConnectedHarwareID": "",
                            "ActiveObjectiveIndex": "0",
                        },
                    ],
                    "Count": "2",
                },
                "Cryostats": {"Count": "0"},
                "Version": "2",
            },
        }
        assert (
            expected_metadata
            == self.s_non_uniform_unfiltered.original_metadata.as_dictionary()
        )


class TestLinescan:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_linescan_path, reader="TriVista", use_uniform_signal_axis=True
        )
        cls.s_non_uniform = hs.load(
            testfile_linescan_path, reader="TriVista", use_uniform_signal_axis=False
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform
        gc.collect()

    def test_data(self):
        expected_data_line1 = np.array([448, 482, 433, 407, 379])
        np.testing.assert_allclose(expected_data_line1, self.s.inav[0].isig[:5].data)
        np.testing.assert_allclose(
            expected_data_line1, self.s_non_uniform.inav[0].isig[:5].data
        )
        expected_data_last_line = np.array([430, 437, 378, 408, 488])
        np.testing.assert_allclose(
            expected_data_last_line, self.s.inav[-1].isig[-5:].data
        )
        np.testing.assert_allclose(
            expected_data_last_line, self.s_non_uniform.inav[-1].isig[-5:].data
        )

    def test_axes(self):
        expected_axis = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "X",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 21,
                "scale": 0.001,
                "offset": -0.01,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 97,
            },
        }

        expected_axis_non_uniform = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "X",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 21,
                "scale": 0.001,
                "offset": -0.01,
            },
            "axis-1": {
                "_type": "DataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
            },
        }
        expected_values_non_uniform_axis = np.array(
            [910.310424804688, 910.350219726563, 910.390014648438]
        )

        uniform_axes_manager = deepcopy(self.s.axes_manager.as_dictionary())
        non_uniform_axis_manager = deepcopy(
            self.s_non_uniform.axes_manager.as_dictionary()
        )

        np.testing.assert_allclose(
            uniform_axes_manager["axis-1"].pop("scale"), 0.0397771, atol=0.0000005
        )
        np.testing.assert_allclose(
            uniform_axes_manager["axis-1"].pop("offset"), 910.3104, atol=0.005
        )
        np.testing.assert_allclose(
            expected_values_non_uniform_axis,
            non_uniform_axis_manager["axis-1"].pop("axis")[:3],
        )
        assert expected_axis == uniform_axes_manager
        assert expected_axis_non_uniform == non_uniform_axis_manager

    def test_original_metadata(self):
        assert (
            self.s.original_metadata.as_dictionary()
            == self.s_non_uniform.original_metadata.as_dictionary()
        )

        metadata_experiment = (
            self.s.original_metadata.Document.InfoSerialized.Experiment.as_dictionary()
        )

        assert metadata_experiment["Mode"] == "Point by Point"
        assert metadata_experiment["Stage Mode"] == "LineScanX"
        assert metadata_experiment["Used Time"] == "00:00:21"


class TestMap:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_map_path, reader="TriVista", use_uniform_signal_axis=True
        )
        cls.s_non_uniform = hs.load(
            testfile_map_path, reader="TriVista", use_uniform_signal_axis=False
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform
        gc.collect()

    def test_data(self):
        expected_data_00 = np.array([531, 499, 442, 446, 460])
        np.testing.assert_allclose(expected_data_00, self.s.inav[0, 0].isig[:5].data)
        np.testing.assert_allclose(
            expected_data_00, self.s_non_uniform.inav[0, 0].isig[:5].data
        )

        expected_data_10 = np.array([532, 516, 425, 438, 433])
        np.testing.assert_allclose(expected_data_10, self.s.inav[1, 0].isig[:5].data)
        np.testing.assert_allclose(
            expected_data_10, self.s_non_uniform.inav[1, 0].isig[:5].data
        )

        expected_data_01 = np.array([525, 498, 447, 433, 452])
        np.testing.assert_allclose(expected_data_01, self.s.inav[0, 1].isig[:5].data)
        np.testing.assert_allclose(
            expected_data_01, self.s_non_uniform.inav[0, 1].isig[:5].data
        )

        expected_data_99 = np.array([452, 478, 687, 1320, 1145])
        np.testing.assert_allclose(expected_data_99, self.s.inav[-1, -1].isig[-5:].data)
        np.testing.assert_allclose(
            expected_data_99, self.s_non_uniform.inav[-1, -1].isig[-5:].data
        )

    def test_axes(self):
        expected_axis = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Y",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 9,
                "scale": 0.025,
                "offset": -0.1,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "X",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 9,
                "scale": 0.025,
                "offset": -0.1,
            },
            "axis-2": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 1024,
            },
        }

        expected_axis_non_uniform = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Y",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 9,
                "scale": 0.025,
                "offset": -0.1,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "X",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 9,
                "scale": 0.025,
                "offset": -0.1,
            },
            "axis-2": {
                "_type": "DataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
            },
        }
        expected_values_non_uniform_axis = np.array(
            [891.894470214844, 891.9345703125, 891.974670410156]
        )

        uniform_axes_manager = deepcopy(self.s.axes_manager.as_dictionary())
        non_uniform_axis_manager = deepcopy(
            self.s_non_uniform.axes_manager.as_dictionary()
        )

        np.testing.assert_allclose(
            uniform_axes_manager["axis-2"].pop("scale"), 0.039770, atol=0.000005
        )
        np.testing.assert_allclose(
            uniform_axes_manager["axis-2"].pop("offset"), 891.9, atol=0.05
        )
        np.testing.assert_allclose(
            expected_values_non_uniform_axis,
            non_uniform_axis_manager["axis-2"].pop("axis")[:3],
        )
        assert expected_axis == uniform_axes_manager
        assert expected_axis_non_uniform == non_uniform_axis_manager

    def test_original_metadata(self):
        assert (
            self.s.original_metadata.as_dictionary()
            == self.s_non_uniform.original_metadata.as_dictionary()
        )

        metadata_experiment = (
            self.s.original_metadata.Document.InfoSerialized.Experiment.as_dictionary()
        )

        assert metadata_experiment["Mode"] == "Point by Point"
        assert metadata_experiment["Stage Mode"] == "MappingXY"
        assert metadata_experiment["Used Time"] == "00:01:27"


class Test3Spectrometers:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_triple_add_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        gc.collect()

    def test_original_metadata(self):
        expected_metadata = {
            "XmlMain": {
                "Version": "1.0.1",
                "Filename": "D:\\Data\\test-spectra\\spec_triple-add.tvf",
                "DateTime": "06/27/2022 17:56:15",
                "PlugInName": "",
                "PlugInData": "",
                "FileInfoSerialized": {"Info": {"Groups": None}},
            },
            "Document": {
                "Version": "2",
                "Label": "Intensity",
                "DataLabel": "Counts",
                "DocType": "Spectra",
                "RecordTime": "06/27/2022 17:55:55.294",
                "ModeName": "",
                "DataSource": "nothing",
                "Encoding": "Ascii",
                "ColorMode": "Grayscale",
                "ViewDisplayMode": "Graph",
                "ViewImageColorMode": "False",
                "InfoSerialized": {
                    "Document": {"Record Time": "6/27/2022 5:55 PM"},
                    "Experiment": {"Used Setup": "PL_add777"},
                    "Spectrometers": {
                        "Spectrometer2": {
                            "Serialnumber": "25580420",
                            "Model": "SP-2-500i",
                            "Stage_Number": "2",
                            "Focallength": "500",
                            "Inclusion_Angle": "8.6",
                            "Detector_Angle": "0",
                            "Groove_Density": "750 g/mm",
                            "Order": "1",
                            "Slit_Entrance-Front": "0",
                            "Slit_Entrance-Side": "6000",
                            "Slit_Exit-Front": "0",
                            "Slit_Exit-Side": "0",
                        },
                        "Spectrometer3": {
                            "Serialnumber": "27580185",
                            "Model": "SP-2-750i",
                            "Stage_Number": "3",
                            "Focallength": "749",
                            "Inclusion_Angle": "6.5",
                            "Detector_Angle": "0.68",
                            "Groove_Density": "750 g/mm",
                            "Order": "1",
                            "Slit_Entrance-Front": "0",
                            "Slit_Entrance-Side": "12000",
                            "Slit_Exit-Front": "0",
                            "Slit_Exit-Side": "0",
                        },
                        "Spectrometer1": {
                            "Serialnumber": "25580419",
                            "Model": "SP-2-500i",
                            "Stage_Number": "1",
                            "Focallength": "500",
                            "Inclusion_Angle": "8.6",
                            "Detector_Angle": "0",
                            "Groove_Density": "750 g/mm",
                            "Order": "1",
                            "Slit_Entrance-Front": "300",
                            "Slit_Entrance-Side": "0",
                            "Slit_Exit-Front": "0",
                            "Slit_Exit-Side": "0",
                        },
                    },
                    "Detector": {
                        "Name": "Camera1",
                        "Serialnumber": None,
                        "Detector_Size": "1024;1",
                        "Detector_Temperature": "-25",
                        "Exposure_Time_(ms)": "1000",
                        "Exposure_Mode": None,
                        "No_of_Accumulations": "1",
                        "Calc_Average": "True",
                        "No_of_Frames": "1",
                        "ADC__Readout_Port": "Normal",
                        "ADC__Rate_Resolution": "1 MHz",
                        "ADC__Gain": "2",
                        "Clearing__Mode": None,
                        "Clearing__No_of_Cleans": "1",
                        "Region_of_Interests": "1|1;1024;1;1;1;1",
                    },
                    "Calibration": {
                        "Center_Wavelength": "820.000",
                        "Laser_Wavelength": "0.000",
                    },
                },
            },
            "Hardware": {
                "Spectrometers": {
                    "Count": "3",
                    "Spectrometer1": {
                        "Gratings": {
                            "Grating": {
                                "Offsets": {"Count": "0"},
                                "GrooveDensity": "750 g/mm",
                                "Blaze": "H-NIR",
                                "Turret": "1",
                            },
                            "Count": "3",
                        },
                        "MirrorEntrance": {
                            "Name": "M1",
                            "State": "isMotorized",
                            "Position": "front",
                        },
                        "MirrorExit": {
                            "Name": "M2",
                            "State": "isMotorized",
                            "Position": "side",
                        },
                        "SlitEntranceSide": {
                            "Name": "S1",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitEntranceFront": {
                            "Name": "S2",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "200",
                        },
                        "SlitExitFront": {
                            "Name": "S3",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitExitSide": {
                            "Name": "S4",
                            "State": "disabled",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "FilterWheel": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "0",
                            "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                        },
                        "Shutter": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "open",
                            "ShutterMode": "OpenForExperiment",
                            "ShutterClosedForFilterwheel": "True",
                            "ShutterClosedForSetupChange": "True",
                            "ShutterClosedForLaserCrossing": "True",
                            "ShutterClosedForLaserCrossingValue": "0.000",
                            "ShutterClosedForLaserCrossingUnit": "Nanometer",
                        },
                        "GUID": "c05fbec6-7cc0-4546-a009-06850868f43d",
                        "Model": "SP-2-500i",
                        "Serialnumber": "25580419",
                        "FocalLength": "500",
                        "InclusionAngle": "8.600",
                        "InclusionAngleAdditive": "51.600",
                        "DetectorAngle": "0.000",
                        "PixelWidthExitFront": "-1.000",
                        "PixelWidthExitSide": "-1.000",
                        "Backlash": "ScanUp",
                        "DriveDirection": "Normal",
                        "ComPortSpec": "COM3",
                        "ComPortEntrance": "",
                        "ComPortExit": "",
                        "DemoMode": "False",
                        "UsedAsLightsource": "False",
                        "UseBackwards": "False",
                        "Wavelength": "925.010",
                        "Grating": "2",
                        "ReadTimeout": "35000",
                        "Activated": "False",
                        "IsFixedCalibration": "False",
                        "FixedCalibrationPoints": "0",
                    },
                    "Spectrometer2": {
                        "Gratings": {
                            "Grating": {
                                "Offsets": {"Count": "0"},
                                "GrooveDensity": "750 g/mm",
                                "Blaze": "H-NIR",
                                "Turret": "1",
                            },
                            "Count": "3",
                        },
                        "MirrorEntrance": {
                            "Name": "M1",
                            "State": "isMotorized",
                            "Position": "side",
                        },
                        "MirrorExit": {
                            "Name": "M2",
                            "State": "isMotorized",
                            "Position": "side",
                        },
                        "SlitEntranceSide": {
                            "Name": "S1",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "10000",
                        },
                        "SlitEntranceFront": {
                            "Name": "S2",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitExitFront": {
                            "Name": "S3",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitExitSide": {
                            "Name": "S4",
                            "State": "disabled",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "FilterWheel": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "0",
                            "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                        },
                        "Shutter": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "open",
                            "ShutterMode": "OpenForExperiment",
                            "ShutterClosedForFilterwheel": "True",
                            "ShutterClosedForSetupChange": "True",
                            "ShutterClosedForLaserCrossing": "True",
                            "ShutterClosedForLaserCrossingValue": "0.000",
                            "ShutterClosedForLaserCrossingUnit": "Nanometer",
                        },
                        "GUID": "39ade9ad-caa8-48c2-a3d0-4f79cc44bda7",
                        "Model": "SP-2-500i",
                        "Serialnumber": "25580420",
                        "FocalLength": "500",
                        "InclusionAngle": "8.600",
                        "InclusionAngleAdditive": "51.600",
                        "DetectorAngle": "0.000",
                        "PixelWidthExitFront": "1.000",
                        "PixelWidthExitSide": "-1.000",
                        "Backlash": "ScanUp",
                        "DriveDirection": "Normal",
                        "ComPortSpec": "COM4",
                        "ComPortEntrance": "",
                        "ComPortExit": "",
                        "DemoMode": "False",
                        "UsedAsLightsource": "False",
                        "UseBackwards": "False",
                        "Wavelength": "-924.910",
                        "Grating": "2",
                        "ReadTimeout": "35000",
                        "Activated": "False",
                        "IsFixedCalibration": "False",
                        "FixedCalibrationPoints": "0",
                    },
                    "Spectrometer3": {
                        "Gratings": {
                            "Grating": {
                                "Offsets": {"Count": "0"},
                                "GrooveDensity": "750 g/mm",
                                "Blaze": "H-NIR",
                                "Turret": "1",
                            },
                            "Count": "3",
                        },
                        "MirrorEntrance": {
                            "Name": "M1",
                            "State": "isMotorized",
                            "Position": "front",
                        },
                        "MirrorExit": {
                            "Name": "M2",
                            "State": "isMotorized",
                            "Position": "side",
                        },
                        "SlitEntranceSide": {
                            "Name": "S1",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitEntranceFront": {
                            "Name": "S2",
                            "State": "isMotorized",
                            "MaxWidth": "12000",
                            "Position": "500",
                        },
                        "SlitExitFront": {
                            "Name": "S3",
                            "State": "disabled",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "SlitExitSide": {
                            "Name": "S4",
                            "State": "disabled",
                            "MaxWidth": "12000",
                            "Position": "0",
                        },
                        "FilterWheel": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "0",
                            "FilterStrings": "6|#1|#2|#3|#4|#5|#6",
                        },
                        "Shutter": {
                            "Name": "",
                            "State": "disabled",
                            "Position": "open",
                            "ShutterMode": "OpenForExperiment",
                            "ShutterClosedForFilterwheel": "True",
                            "ShutterClosedForSetupChange": "True",
                            "ShutterClosedForLaserCrossing": "True",
                            "ShutterClosedForLaserCrossingValue": "0.000",
                            "ShutterClosedForLaserCrossingUnit": "Nanometer",
                        },
                        "GUID": "37a2855b-db41-4268-970b-bb75d96e5654",
                        "Model": "SP-2-750i",
                        "Serialnumber": "27580185",
                        "FocalLength": "749",
                        "InclusionAngle": "6.500",
                        "InclusionAngleAdditive": "39.000",
                        "DetectorAngle": "0.680",
                        "PixelWidthExitFront": "25.000",
                        "PixelWidthExitSide": "-1.000",
                        "Backlash": "ScanUp",
                        "DriveDirection": "Normal",
                        "ComPortSpec": "COM5",
                        "ComPortEntrance": "",
                        "ComPortExit": "",
                        "DemoMode": "False",
                        "UsedAsLightsource": "False",
                        "UseBackwards": "False",
                        "Wavelength": "850.184",
                        "Grating": "2",
                        "ReadTimeout": "35000",
                        "Activated": "False",
                        "IsFixedCalibration": "False",
                        "FixedCalibrationPoints": "0",
                    },
                },
                "Detectors": {
                    "Detector": {
                        "GUID": "c1bc9fd2-135e-4098-b681-d30f6ea8be82",
                        "Driver": "PvCam32 Driver",
                        "Name": "Camera1",
                        "DisplayName": "Camera1",
                        "SerialNumber": "",
                        "DetectorType": "None",
                        "Size": "1024;1",
                        "OpenAtStartup": "False",
                        "Usage": "Default",
                    },
                    "Count": "3",
                },
                "SinglePointDetectors": {"Count": "0"},
                "LightSources": {
                    "LightSource": [
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "0.000"},
                            "GUID": "a57ac71a-c667-4501-a24a-596bc54b2c8c",
                            "Name": "Calibration Lamp",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "0"},
                            "GUID": "cac03351-735e-4592-9bb2-c17f64b3649f",
                            "Name": "Calibration Lamp2",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "0"},
                            "GUID": "84a08bee-227a-46f0-b1e1-0f171808a155",
                            "Name": "Calibration Lamp3",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "False",
                            "Type": "CalibrationLamp",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "0.000"},
                            "GUID": "a0601d22-23a0-4587-bb79-35d5823916c4",
                            "Name": "Laser",
                            "Serialnumber": "",
                            "FixedWavelength": "False",
                            "CanChange": "False",
                            "Type": "Laser",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                        {
                            "Wavelengths": {"Count": "1", "Value_0": "790.000"},
                            "GUID": "a36b29da-0d54-4ab7-860c-967269f76217",
                            "Name": "tisa",
                            "Serialnumber": "",
                            "FixedWavelength": "True",
                            "CanChange": "True",
                            "Type": "Laser",
                            "ConnectionString": "None",
                            "IsMirrorPositionValid": "False",
                            "MirrorsAlignmentPositionTop": "0;0",
                            "MirrorsAlignmentPositionBottom": "0;0",
                        },
                    ],
                    "Count": "5",
                },
                "ArduinoHardware": {"Count": "0"},
                "Additionals": {"Count": "0"},
                "MotorizedStages": {
                    "MotorizedStage": {
                        "GUID": "c551372e-56af-4686-b7c1-48ca00dfdf27",
                        "DriverName": "Tango Driver",
                        "Name": "XY-Stage",
                        "ConnectionString": "COM15",
                        "Serialnumber": "144012056",
                        "IsTriggerModeAvailable": "False",
                    },
                    "Count": "1",
                },
                "Microscopes": {
                    "Microscope": {
                        "Objectives": {
                            "Objective": {
                                "IsEnabled": "True",
                                "Name": "100x",
                                "Magnification": "100.000",
                                "FactorMeasuredWithResolution": "1280",
                                "Factor": "34.700",
                                "LaserOffsetX": "0.000",
                                "LaserOffsetY": "0.000",
                                "OffsetX": "0.000",
                                "OffsetY": "0.000",
                                "LineLaserOffsetX": "0.000",
                                "LineLaserOffsetY": "0.000",
                                "LineOffsetX": "0.000",
                                "LineOffsetY": "0.000",
                                "LineFocusFactor": "1.130",
                            },
                            "Count": "10",
                        },
                        "GUID": "841bfef9-b969-43f0-9ca3-2522c223a31f",
                        "Name": "Microscope",
                        "StageID": "c551372e-56af-4686-b7c1-48ca00dfdf27",
                        "ImagingCameraID": "",
                        "RamanBoxAvailable": "True",
                        "RamanBoxLinearStage.Available": "False",
                        "RamanBoxLinearStage.ConnectedHarwareID": "",
                        "RamanBoxFilterWheel.Available": "False",
                        "RamanBoxFilterWheel.ConnectedHarwareID": "",
                        "RamanBoxBeamSplitter.Available": "False",
                        "RamanBoxBeamSplitter.ConnectedHarwareID": "",
                        "RamanBoxNotchFilter.Available": "False",
                        "RamanBoxNotchFilter.ConnectedHarwareID": "",
                        "ActiveObjectiveIndex": "0",
                    },
                    "Count": "2",
                },
                "Cryostats": {"Count": "0"},
                "Version": "2",
            },
        }

        assert self.s.original_metadata.as_dictionary() == expected_metadata

    def test_metadata_acquisition_instrument(self):
        expected_metadata = {
            "Detector": {
                "processing": {"calc_average": "True"},
                "temperature": -25.0,
                "exposure_per_frame": 1.0,
                "frames": 1.0,
                "glued_spectrum": False,
                "integration_time": 1.0,
            },
            "Laser": {"objective_magnification": 100.0},
            "Spectrometer1": {
                "Grating": {"blazing_wavelength": "H-NIR", "groove_density": 750.0},
                "central_wavelength": 820.0,
                "model": "SP-2-500i",
                "entrance_slit_width": 0.3,
                "exit_slit_width": 0.0,
            },
            "Spectrometer2": {
                "Grating": {"blazing_wavelength": "H-NIR", "groove_density": 750.0},
                "central_wavelength": 820.0,
                "model": "SP-2-500i",
                "entrance_slit_width": 6.0,
                "exit_slit_width": 0.0,
            },
            "Spectrometer3": {
                "Grating": {"blazing_wavelength": "H-NIR", "groove_density": 750.0},
                "central_wavelength": 820.0,
                "model": "SP-2-750i",
                "entrance_slit_width": 12.0,
                "exit_slit_width": 0.0,
            },
        }

        assert (
            self.s.metadata.Acquisition_instrument.as_dictionary() == expected_metadata
        )


class TestStepAndGlue:
    @classmethod
    def setup_class(cls):
        cls.glued = hs.load(
            testfile_step_and_glue_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
            glued_data_as_stack=False,
        )
        cls.stack = hs.load(
            testfile_step_and_glue_path,
            reader="TriVista",
            use_uniform_signal_axis=False,
            filter_original_metadata=True,
            glued_data_as_stack=True,
        )

    @classmethod
    def teardown_class(cls):
        del cls.glued
        del cls.stack
        gc.collect()

    def test_data_glued(self):
        assert np.isclose(self.glued.isig[0].data, 1071)
        assert np.isclose(self.glued.isig[17999].data, 978)

    def test_data_stack(self):
        data_stack1 = np.array([1071, 1134, 1092])
        data_stack19 = np.array([1064, 1150, 1111])
        np.testing.assert_allclose(self.stack[0].isig[:3].data, data_stack1)
        np.testing.assert_allclose(self.stack[18].isig[:3].data, data_stack19)

    def test_axes_stack(self):
        expected_axis = {
            "axis-0": {
                "_type": "DataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
            }
        }
        for frame in self.stack:
            axes_manager = frame.axes_manager.as_dictionary()
            data_axis = axes_manager["axis-0"].pop("axis")
            assert data_axis.size == 1024
            assert axes_manager == expected_axis
        assert np.isclose(
            self.stack[0].axes_manager.as_dictionary()["axis-0"]["axis"][0],
            900.027,
            atol=0.001,
        )

    def test_metadata(self):
        original_metadata_glued = self.glued.original_metadata.Document.InfoSerialized.Experiment.as_dictionary()  # noqa: E501

        metadata_detector = self.glued.metadata.Acquisition_instrument.Detector

        assert original_metadata_glued["From"] == "900.000 nm"
        assert original_metadata_glued["To"] == "1500.000 nm"
        assert original_metadata_glued["Mode:"] == "Step & Glue"
        assert original_metadata_glued["Overlap (%)"] == "15"
        assert original_metadata_glued["Skipped Pixel Left"] == "0"
        assert original_metadata_glued["Skipped Pixel Right"] == "0"

        assert metadata_detector.glued_spectrum is True
        assert np.isclose(metadata_detector.glued_spectrum_overlap, 15)
        assert np.isclose(metadata_detector.glued_spectrum_windows, 19)

        metadata_glued = deepcopy(self.glued.metadata.as_dictionary())
        del metadata_glued["General"]["FileIO"]
        for frame in self.stack:
            assert (
                self.glued.original_metadata.as_dictionary()
                == frame.original_metadata.as_dictionary()
            )
            metadata_frame = deepcopy(frame.metadata.as_dictionary())
            del metadata_frame["General"]["FileIO"]
            assert metadata_frame == metadata_glued


class TestTimeSeries:
    @classmethod
    def setup_class(cls):
        cls.timeseries = hs.load(
            testfile_spec_timeseries_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
        )
        cls.frames = hs.load(
            testfile_spec_2frames_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
        )

    @classmethod
    def teardown_class(cls):
        del cls.timeseries
        del cls.frames
        gc.collect()

    def test_data_timeseries(self):
        data_ts_1 = np.array([27381, 39153, 35022, -425, 32430])
        data_ts_2 = np.array([27449.5, 38971.5, 35168.5, -425, 32484.5])
        np.testing.assert_allclose(self.timeseries.isig[:5, 0].data, data_ts_1)
        np.testing.assert_allclose(self.timeseries.isig[:5, 1].data, data_ts_2)

    def test_data_frames(self):
        data_frame_1 = np.array([49868, 65074, 60325, 10241, 52539])
        data_frame_2 = np.array([49909, 65074, 60017, 10022, 52639])
        np.testing.assert_allclose(self.frames.isig[:5, 0].data, data_frame_1)
        np.testing.assert_allclose(self.frames.isig[:5, 1].data, data_frame_2)

    def test_axes_timeseries(self):
        expected_axis = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "time",
                "units": "s",
                "navigate": False,
                "is_binned": False,
                "size": 2,
                "offset": 0.0,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 1024,
            },
        }

        axes_manager = deepcopy(self.timeseries.axes_manager.as_dictionary())

        np.testing.assert_allclose(axes_manager["axis-0"].pop("scale"), 4.04, atol=0.01)
        np.testing.assert_allclose(axes_manager["axis-1"].pop("scale"), 0.03, atol=0.01)
        np.testing.assert_allclose(axes_manager["axis-1"].pop("offset"), 934, atol=1)
        assert expected_axis == axes_manager

    def test_axes_frames(self):
        expected_axis = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "time",
                "units": "s",
                "navigate": False,
                "is_binned": False,
                "size": 2,
                "offset": 0.0,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 1024,
            },
        }

        axes_manager = deepcopy(self.frames.axes_manager.as_dictionary())

        np.testing.assert_allclose(axes_manager["axis-1"].pop("scale"), 0.03, atol=0.01)
        np.testing.assert_allclose(axes_manager["axis-1"].pop("offset"), 934, atol=1)
        np.testing.assert_allclose(axes_manager["axis-0"].pop("scale"), 6, atol=0.01)
        assert expected_axis == axes_manager


class TestSpecIntegrationTime:
    @classmethod
    def setup_class(cls):
        cls.s_2acc = hs.load(
            testfile_spec_2acc_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
            glued_data_as_stack=True,
        )
        ## glued_data_as_stack set to True here
        ## to ensure that this setting doesn't affect
        ## non-glued datasets
        cls.s_2acc_no_average = hs.load(
            testfile_spec_2acc_no_average_path,
            reader="TriVista",
            use_uniform_signal_axis=True,
            filter_original_metadata=True,
            glued_data_as_stack=True,
        )

    @classmethod
    def teardown_class(cls):
        del cls.s_2acc
        del cls.s_2acc_no_average
        gc.collect()

    def test_metadata_2acc(self):
        metadata_detector = self.s_2acc.metadata.Acquisition_instrument.Detector
        assert np.isclose(metadata_detector.exposure_per_frame, 3)
        assert np.isclose(metadata_detector.frames, 2)
        assert np.isclose(metadata_detector.integration_time, 3)
        assert metadata_detector.processing.calc_average == "True"

    def test_metadata_2acc_no_average(self):
        metadata_detector = (
            self.s_2acc_no_average.metadata.Acquisition_instrument.Detector
        )
        assert np.isclose(metadata_detector.exposure_per_frame, 3)
        assert np.isclose(metadata_detector.frames, 2)
        assert np.isclose(metadata_detector.integration_time, 6)
        assert metadata_detector.processing.calc_average == "False"



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
#
"""Utility functions for HDF5 file inspection."""

import json
import pprint

import h5py

from rsciio._docstrings import FILENAME_DOC, LAZY_DOC
from rsciio.nexus._api import (
    _check_search_keys,
    _find_data,
    _find_search_keys_in_dict,
    _load_metadata,
)

__all__ = [
    "list_datasets_in_file",
    "read_metadata_from_file",
]


def __dir__():
    return sorted(__all__)


def read_metadata_from_file(
    filename, lazy=False, metadata_key=None, verbose=False, skip_array_metadata=False
):
    """
    Read the metadata from a NeXus or ``.hdf`` file.

    This method iterates through the hdf5 file and returns a dictionary of
    the entries.
    This is a convenience method to inspect a file for a value
    rather than loading the file as a signal.

    Parameters
    ----------
    %s
    %s
    metadata_key : None, str, list of str, default=None
        None will return all datasets found including linked data.
        Providing a string or list of strings will only return items
        which contain the string(s).
        For example, search_keys = ["instrument","Fe"] will return
        hdf entries with "instrument" or "Fe" in their hdf path.
    verbose : bool, default=False
        Pretty print the results to screen.
    skip_array_metadata : bool, default=False
        Whether to skip loading array metadata. This is useful as a lot of
        large array may be present in the metadata and it is redundant with
        dataset itself.

    Returns
    -------
    dict
        Metadata dictionary.

    See Also
    --------
    rsciio.utils.hdf5.list_datasets_in_file : Convenience function to list
        datasets present in a file.
    """
    search_keys = _check_search_keys(metadata_key)
    fin = h5py.File(filename, "r")
    # search for NXdata sets...
    # strip out the metadata (basically everything other than NXdata)
    stripped_metadata = _load_metadata(
        fin, lazy=lazy, skip_array_metadata=skip_array_metadata
    )
    stripped_metadata = _find_search_keys_in_dict(
        stripped_metadata, search_keys=search_keys
    )
    if verbose:
        pprint.pprint(stripped_metadata)  # noqa: T203

    fin.close()
    return stripped_metadata


read_metadata_from_file.__doc__ %= (FILENAME_DOC, LAZY_DOC)


def list_datasets_in_file(
    filename, dataset_key=None, hardlinks_only=False, verbose=True
):
    """
    Read from a NeXus or ``.hdf`` file and return a list of the dataset paths.

    This method is used to inspect the contents of an hdf5 file.
    The method iterates through group attributes and returns NXdata or
    hdf datasets of size >=2 if they're not already NXdata blocks
    and returns a list of the entries.
    This is a convenience method to inspect a file to list datasets
    present rather than loading all the datasets in the file as signals.

    Parameters
    ----------
    %s
    dataset_key : str, list of str, None , default=None
        If a str or list of strings is provided only return items whose
        path contain the strings.
        For example, dataset_key = ["instrument", "Fe"] will only return
        hdf entries with "instrument" or "Fe" somewhere in their hdf path.
    hardlinks_only : bool, default=False
        If true any links (soft or External) will be ignored when loading.
    verbose : bool, default=True
        Prints the results to screen.

    Returns
    -------
    list
        List of paths to datasets.

    See Also
    --------
    rsciio.utils.hdf5.read_metadata_from_file : Convenience function to read
        metadata present in a file.
    """
    search_keys = _check_search_keys(dataset_key)
    fin = h5py.File(filename, "r")
    # search for NXdata sets...
    # strip out the metadata (basically everything other than NXdata)
    nexus_data_paths, hdf_dataset_paths = _find_data(
        fin, search_keys=search_keys, hardlinks_only=hardlinks_only
    )
    if verbose:
        if nexus_data_paths:
            print("NXdata found")  # noqa: T201
            for nxd in nexus_data_paths:
                print(nxd)  # noqa: T201
        else:
            print("NXdata not found")  # noqa: T201
        if hdf_dataset_paths:
            print("Unique HDF datasets found")  # noqa: T201
            for hdfd in hdf_dataset_paths:
                print(hdfd, fin[hdfd].shape)  # noqa: T201
        else:
            print("No HDF datasets found or data is captured by NXdata")  # noqa: T201
    fin.close()
    return nexus_data_paths, hdf_dataset_paths


list_datasets_in_file.__doc__ %= FILENAME_DOC


def _get_keys_from_group(group):
    # Return a list of ids of items contains in the group
    return list(group.keys())


def _parse_sub_data_group_metadata(sub_data_group):
    metadata_array = sub_data_group["Metadata"][:, 0].T
    mdata_string = metadata_array.tobytes().decode("utf-8")
    return json.loads(mdata_string.rstrip("\x00"))


def _parse_metadata(data_group, sub_group_key):
    return _parse_sub_data_group_metadata(data_group[sub_group_key])



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import gc
import warnings
from pathlib import Path

import dask.array as da
import numpy as np
import pytest

from rsciio.utils._array import sarray2dict
from rsciio.utils._date_time import serial_date_to_ISO_format
from rsciio.utils._tests import assert_deep_almost_equal

try:
    WindowsError
except NameError:
    WindowsError = None

pytest.importorskip("skimage", reason="scikit-image not installed")
hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")

from rsciio.blockfile._api import get_default_header  # noqa: E402

TEST_DATA_DIR = Path(__file__).parent / "data" / "blockfile"
FILE1 = TEST_DATA_DIR / "test1.blo"
FILE2 = TEST_DATA_DIR / "test2.blo"


@pytest.fixture()
def fake_signal():
    fake_data = np.arange(300, dtype=np.uint8).reshape(3, 4, 5, 5)
    fake_signal = hs.signals.Signal2D(fake_data)
    fake_signal.axes_manager[0].scale_as_quantity = "1 mm"
    fake_signal.axes_manager[1].scale_as_quantity = "1 mm"
    fake_signal.axes_manager[2].scale_as_quantity = "1 mm"
    fake_signal.axes_manager[3].scale_as_quantity = "1 mm"
    return fake_signal


@pytest.fixture()
def save_path(tmp_path):
    filepath = tmp_path / "save_temp.blo"
    yield filepath
    # Force files release (required in Windows)
    gc.collect()


ref_data2 = np.array(
    [
        [
            [
                [20, 23, 25, 25, 27],
                [29, 23, 23, 0, 29],
                [24, 0, 0, 22, 18],
                [0, 14, 19, 17, 26],
                [19, 21, 22, 27, 20],
            ],
            [
                [28, 25, 29, 15, 29],
                [12, 15, 12, 25, 24],
                [25, 26, 26, 18, 27],
                [19, 18, 20, 23, 28],
                [28, 18, 22, 25, 0],
            ],
            [
                [21, 29, 25, 19, 18],
                [30, 15, 20, 22, 26],
                [23, 18, 26, 15, 25],
                [22, 25, 24, 15, 20],
                [22, 15, 15, 21, 23],
            ],
        ],
        [
            [
                [28, 25, 26, 24, 26],
                [26, 17, 0, 24, 12],
                [17, 18, 21, 19, 21],
                [21, 24, 19, 17, 0],
                [17, 14, 25, 15, 26],
            ],
            [
                [25, 18, 20, 15, 24],
                [19, 13, 23, 18, 11],
                [0, 25, 0, 0, 14],
                [26, 22, 22, 11, 14],
                [21, 0, 15, 13, 19],
            ],
            [
                [24, 18, 20, 22, 21],
                [13, 25, 20, 28, 29],
                [15, 17, 24, 23, 23],
                [22, 21, 21, 22, 18],
                [24, 25, 18, 18, 27],
            ],
        ],
    ],
    dtype=np.uint8,
)

axes1 = {
    "axis-0": {
        "_type": "UniformDataAxis",
        "name": "y",
        "navigate": True,
        "offset": 0.0,
        "scale": 12.8,
        "size": 3,
        "units": "nm",
        "is_binned": False,
    },
    "axis-1": {
        "_type": "UniformDataAxis",
        "name": "x",
        "navigate": True,
        "offset": 0.0,
        "scale": 12.8,
        "size": 2,
        "units": "nm",
        "is_binned": False,
    },
    "axis-2": {
        "_type": "UniformDataAxis",
        "name": "dy",
        "navigate": False,
        "offset": 0.0,
        "scale": 0.016061676839061997,
        "size": 144,
        "units": "cm",
        "is_binned": False,
    },
    "axis-3": {
        "_type": "UniformDataAxis",
        "name": "dx",
        "navigate": False,
        "offset": 0.0,
        "scale": 0.016061676839061997,
        "size": 144,
        "units": "cm",
        "is_binned": False,
    },
}

axes2 = {
    "axis-0": {
        "_type": "UniformDataAxis",
        "name": "y",
        "navigate": True,
        "offset": 0.0,
        "scale": 64.0,
        "size": 2,
        "units": "nm",
        "is_binned": False,
    },
    "axis-1": {
        "_type": "UniformDataAxis",
        "name": "x",
        "navigate": True,
        "offset": 0.0,
        "scale": 64.0,
        "size": 3,
        "units": "nm",
        "is_binned": False,
    },
    "axis-2": {
        "_type": "UniformDataAxis",
        "name": "dy",
        "navigate": False,
        "offset": 0.0,
        "scale": 0.016061676839061997,
        "size": 5,
        "units": "cm",
        "is_binned": False,
    },
    "axis-3": {
        "_type": "UniformDataAxis",
        "name": "dx",
        "navigate": False,
        "offset": 0.0,
        "scale": 0.016061676839061997,
        "size": 5,
        "units": "cm",
        "is_binned": False,
    },
}

axes2_converted = {
    "axis-0": {
        "_type": "UniformDataAxis",
        "name": "y",
        "navigate": True,
        "offset": 0.0,
        "scale": 64.0,
        "size": 2,
        "units": "nm",
        "is_binned": False,
    },
    "axis-1": {
        "_type": "UniformDataAxis",
        "name": "x",
        "navigate": True,
        "offset": 0.0,
        "scale": 64.0,
        "size": 3,
        "units": "nm",
        "is_binned": False,
    },
    "axis-2": {
        "_type": "UniformDataAxis",
        "name": "dy",
        "navigate": False,
        "offset": 0.0,
        "scale": 160.61676839061997,
        "size": 5,
        "units": "µm",
        "is_binned": False,
    },
    "axis-3": {
        "_type": "UniformDataAxis",
        "name": "dx",
        "navigate": False,
        "offset": 0.0,
        "scale": 160.61676839061997,
        "size": 5,
        "units": "µm",
        "is_binned": False,
    },
}


def test_load1():
    s = hs.load(FILE1)
    assert s.data.shape == (3, 2, 144, 144)
    assert s.axes_manager.as_dictionary() == axes1


@pytest.mark.parametrize(("convert_units"), (True, False))
def test_load2(convert_units):
    s = hs.load(FILE2, convert_units=convert_units)
    assert s.data.shape == (2, 3, 5, 5)
    axes = axes2_converted if convert_units else axes2
    np.testing.assert_equal(s.axes_manager.as_dictionary(), axes)
    np.testing.assert_allclose(s.data, ref_data2)


@pytest.mark.parametrize(("convert_units"), (True, False))
def test_save_load_cycle(save_path, convert_units):
    sig_reload = None
    signal = hs.load(FILE2, convert_units=convert_units)
    serial = signal.original_metadata["blockfile_header"]["Acquisition_time"]
    date, time, timezone = serial_date_to_ISO_format(serial)
    assert signal.metadata.General.original_filename == "test2.blo"
    assert signal.metadata.General.date == date
    assert signal.metadata.General.time == time
    assert signal.metadata.General.time_zone == timezone
    assert (
        signal.metadata.General.notes
        == "Precession angle : \r\nPrecession Frequency : \r\nCamera gamma : on"
    )
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path, convert_units=convert_units)
    np.testing.assert_equal(signal.data, sig_reload.data)
    assert (
        signal.axes_manager.as_dictionary() == sig_reload.axes_manager.as_dictionary()
    )
    assert (
        signal.original_metadata.as_dictionary()
        == sig_reload.original_metadata.as_dictionary()
    )
    # change original_filename to make the metadata of both signals equals
    sig_reload.metadata.General.original_filename = (
        signal.metadata.General.original_filename
    )
    # assert file reading tests here, then delete so we can compare
    # entire metadata structure at once:
    plugin = "rsciio.blockfile"
    assert signal.metadata.General.FileIO.Number_0.operation == "load"
    assert signal.metadata.General.FileIO.Number_0.io_plugin == plugin
    assert signal.metadata.General.FileIO.Number_1.operation == "save"
    assert signal.metadata.General.FileIO.Number_1.io_plugin == plugin
    assert sig_reload.metadata.General.FileIO.Number_0.operation == "load"
    assert sig_reload.metadata.General.FileIO.Number_0.io_plugin == plugin
    del signal.metadata.General.FileIO
    del sig_reload.metadata.General.FileIO

    assert_deep_almost_equal(
        signal.metadata.as_dictionary(), sig_reload.metadata.as_dictionary()
    )
    assert signal.metadata.General.date == sig_reload.metadata.General.date
    assert signal.metadata.General.time == sig_reload.metadata.General.time
    assert isinstance(signal, hs.signals.Signal2D)
    # Delete reference to close memmap file!
    del sig_reload


def test_different_x_y_scale_units(save_path):
    # perform load and save cycle with changing the scale on y
    signal = hs.load(FILE2)
    signal.axes_manager[0].scale = 50.0
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_allclose(sig_reload.axes_manager[0].scale, 50.0, rtol=1e-5)
    np.testing.assert_allclose(sig_reload.axes_manager[1].scale, 64.0, rtol=1e-5)
    np.testing.assert_allclose(sig_reload.axes_manager[2].scale, 0.0160616, rtol=1e-5)


def test_inconvertible_units(save_path, fake_signal):
    fake_signal.axes_manager[2].units = "1/A"
    fake_signal.axes_manager[3].units = "1/A"
    with pytest.warns(UserWarning):
        fake_signal.save(save_path, overwrite=True)


def test_overflow(save_path, fake_signal):
    fake_signal.change_dtype(np.uint16)
    with pytest.warns(UserWarning):
        fake_signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_allclose(sig_reload.data, fake_signal.data.astype(np.uint8))


def test_dtype_lims(save_path, fake_signal):
    fake_signal.data = fake_signal.data * 100
    fake_signal.change_dtype(np.uint16)
    fake_signal.save(save_path, intensity_scaling="dtype", overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_allclose(
        sig_reload.data, (fake_signal.data / 65535 * 255).astype(np.uint8)
    )


def test_dtype_float_fail(save_path, fake_signal):
    fake_signal.change_dtype(np.float32)
    with pytest.raises(ValueError):
        fake_signal.save(save_path, intensity_scaling="dtype", overwrite=True)


def test_minmax_lims(save_path, fake_signal):
    fake_signal.save(save_path, intensity_scaling="minmax", overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_allclose(
        sig_reload.data,
        (fake_signal.data / fake_signal.data.max() * 255).astype(np.uint8),
    )


def test_crop_lims(save_path, fake_signal):
    fake_signal.save(save_path, intensity_scaling="crop", overwrite=True)
    sig_reload = hs.load(save_path)
    compare = fake_signal.data
    compare[compare > 255] = 255
    np.testing.assert_allclose(sig_reload.data, compare)


def test_tuple_limits(save_path, fake_signal):
    skimage = pytest.importorskip("skimage", reason="scikit-image not installed")
    fake_signal.save(save_path, intensity_scaling=(5, 200), overwrite=True)
    sig_reload = hs.load(save_path)
    compare = skimage.exposure.rescale_intensity(
        fake_signal.data, in_range=(5, 200), out_range=np.uint8
    )
    np.testing.assert_allclose(sig_reload.data, compare)


def test_lazy_save(save_path, fake_signal):
    fake_signal = fake_signal.as_lazy()
    fake_signal.save(save_path, intensity_scaling="minmax", overwrite=True)
    sig_reload = hs.load(save_path)
    compare = (fake_signal.data / fake_signal.data.max() * 255).astype(np.uint8)
    np.testing.assert_allclose(sig_reload.data, compare)


@pytest.mark.parametrize("navigator", [None, "navigator", "array"])
def test_vbfs(save_path, fake_signal, navigator):
    fake_signal = fake_signal.as_lazy()
    if navigator in ["navigator", "array"]:
        fake_signal.compute_navigator()
    if navigator == "array":
        navigator = fake_signal.navigator.data
    fake_signal.save(
        save_path, intensity_scaling=None, navigator=navigator, overwrite=True
    )
    sig_reload = hs.load(save_path)
    np.testing.assert_allclose(sig_reload.data, fake_signal.data)


def test_invalid_vbf(save_path, fake_signal):
    with pytest.raises(ValueError):
        fake_signal.save(
            save_path,
            navigator=hs.signals.Signal2D(np.zeros((10, 10))),
            overwrite=True,
        )


def test_default_header():
    # Simply check that no exceptions are raised
    header = get_default_header()
    assert header is not None


def test_non_square(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 6)).astype(np.uint8))
    with pytest.warns(UserWarning):
        # warning about expect cm units
        with pytest.raises(ValueError):
            signal.save(save_path, overwrite=True)


def test_load_lazy():
    s = hs.load(FILE2, lazy=True)
    assert isinstance(s.data, da.Array)
    s.compute()
    s2 = hs.load(FILE2)
    np.testing.assert_allclose(s.data, s2.data)


def test_load_to_memory():
    s = hs.load(FILE2, lazy=False)
    assert isinstance(s.data, np.ndarray)
    assert not isinstance(s.data, np.memmap)


def test_write_fresh(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 5)).astype(np.uint8))
    signal.axes_manager["sig"].set(units="cm")
    signal.axes_manager["nav"].set(units="nm")
    signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_equal(signal.data, sig_reload.data)
    header = sarray2dict(get_default_header())
    header.update(
        {
            "NX": 3,
            "NY": 10,
            "DP_SZ": 5,
            "SX": 1,
            "SY": 1,
            "SDP": 100,
            "Data_offset_2": 10 * 3 + header["Data_offset_1"],
            "Note": "",
        }
    )
    header["Data_offset_2"] += header["Data_offset_2"] % 16
    assert sig_reload.original_metadata.blockfile_header.as_dictionary() == header


def test_write_data_line(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(3, 5, 5)).astype(np.uint8))
    with pytest.warns(UserWarning):
        # expected units warning
        signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_equal(signal.data, sig_reload.data)


def test_write_data_single(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(5, 5)).astype(np.uint8))
    with pytest.warns(UserWarning):
        # expected units warning
        signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    np.testing.assert_equal(signal.data, sig_reload.data)


def test_write_data_am_mismatch(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 5)).astype(np.uint8))
    signal.axes_manager.navigation_axes[1].size = 4
    with pytest.warns(UserWarning):
        # expected units warning
        with pytest.raises(ValueError):
            signal.save(save_path, overwrite=True)


def test_unrecognized_header_warning(save_path, fake_signal):
    fake_signal.save(save_path, overwrite=True)
    # change magic number
    with open(save_path, "r+b") as f:
        f.seek(6)
        f.write((0xAFAF).to_bytes(2, byteorder="big", signed=False))
    with pytest.warns(UserWarning, match=r"Blockfile has unrecognized .*"):
        hs.load(save_path)


def test_write_cutoff(save_path):
    signal = hs.signals.Signal2D((255 * np.random.rand(10, 3, 5, 5)).astype(np.uint8))
    signal.axes_manager.navigation_axes[0].size = 20
    signal.axes_manager["sig"].set(units="cm")
    signal.axes_manager["nav"].set(units="nm")
    signal.save(save_path, overwrite=True)
    # Test that it raises a warning
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        sig_reload = hs.load(save_path)
        # There can be other warnings so >=
        assert len(w) >= 1
        warning_blockfile = [
            "Blockfile header" in str(warning.message) for warning in w
        ]
        assert True in warning_blockfile
        assert issubclass(w[warning_blockfile.index(True)].category, UserWarning)
    cut_data = signal.data.flatten()
    pw = [(0, 17 * 10 * 5 * 5)]
    cut_data = np.pad(cut_data, pw, mode="constant")
    cut_data = cut_data.reshape((10, 20, 5, 5))
    np.testing.assert_equal(cut_data, sig_reload.data)


def test_crop_notes(save_path):
    note_len = 0x1000 - 0xF0
    note = "test123" * 1000  # > note_len
    signal = hs.signals.Signal2D((255 * np.random.rand(2, 3, 2, 2)).astype(np.uint8))
    signal.original_metadata.add_node("blockfile_header.Note")
    signal.original_metadata.blockfile_header.Note = note
    with pytest.warns(UserWarning):
        # expected units warning
        signal.save(save_path, overwrite=True)
    sig_reload = hs.load(save_path)
    assert sig_reload.original_metadata.blockfile_header.Note == note[:note_len]


def test_blo_chunking_lazy(save_path):
    data = np.zeros((90, 121, 144, 144), dtype=np.uint8)
    sig = hs.signals.Signal2D(data)
    sig.save(save_path, overwrite=True)
    new_s = hs.load(save_path, lazy=True, chunks=(1, 1, 144 // 2, 144 // 2))
    assert isinstance(new_s.data, da.Array)
    assert new_s.data.shape == (90, 121, 144, 144)
    assert new_s.data.chunks[0] == (1,) * 90
    assert new_s.data.chunks[1] == (1,) * 121
    assert new_s.data.chunks[2] == (72, 72)
    assert new_s.data.chunks[3] == (72, 72)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import os
import sys
import warnings
from pathlib import Path

import pooch

PATH = Path(__file__).parent


def update_registry():
    """
    Update the ``rsciio.tests.registry.txt`` file, which is required after
    adding or updating test data files.

    Unix system only. This is not supported on windows, because the hash
    comparison will fail for non-binary file, because of difference in line
    ending.
    """
    if sys.platform == "win32":
        warnings.warn("Updating registry is not supported on Windows. Nothing done.")
        return

    make_registry(
        PATH / "data",
        PATH / "registry.txt",
        recursive=True,
        exclude_pattern=[".git", ".DS_Store"],
    )


def make_registry(directory, output, recursive=True, exclude_pattern=None):
    """
    Make a registry of files and hashes for the given directory.

    This is helpful if you have many files in your test dataset as it keeps you
    from needing to manually update the registry.

    Parameters
    ----------
    directory : str
        Directory of the test data to put in the registry. All file names in
        the registry will be relative to this directory.
    output : str
        Name of the output registry file.
    recursive : bool
        If True, will recursively look for files in subdirectories of
        *directory*.
    exclude_pattern : list or None
        List of pattern to exclude.

    Notes
    -----
    Adapted from https://github.com/fatiando/pooch/blob/main/pooch/hashes.py
    BSD-3-Clause
    """
    directory = Path(directory)
    if recursive:
        pattern = "**/*"
    else:
        pattern = "*"

    if exclude_pattern is None:
        exclude_pattern = []

    files = set(
        str(path.relative_to(directory))
        for path in directory.glob(pattern)
        if path.is_file()
    )

    files_to_exclude = set(
        file for file in files if any(pattern in file for pattern in exclude_pattern)
    )

    files = sorted(files - files_to_exclude)

    hashes = [pooch.file_hash(str(directory / fname)) for fname in files]

    with open(output, "w") as outfile:
        for fname, fhash in zip(files, hashes):
            # Only use Unix separators for the registry so that we don't go
            # insane dealing with file paths.
            outfile.write("'{}' {}\n".format(fname.replace("\\", "/"), fhash))


def download_all(pooch_object=None, ignore_hash=None, show_progressbar=True):
    """
    Download all test data if they are not already locally available in
    ``rsciio.tests.data`` folder.

    When the environment variable `RSCIIO_TEST_DATA_SOURCE` is set,
    the test data will be copied from the specified directory instead of downloading.
    This is useful for CI when the test data is cached as an artifact in a previous step.

    Parameters
    ----------
    pooch_object : pooch.Pooch or None, default=None
        The registry to be used. If None, a RosettaSciIO registry will
        be used.
    ignore_hash : bool or None, default=None
        Don't compare the hash of the downloaded file with the corresponding
        hash in the registry. On windows, the hash comparison will fail for
        non-binary file, because of difference in line ending. If None, the
        comparison will only be used on unix system.
    show_progressbar : bool, default=True
        Whether to show the progressbar or not.
    """

    src = os.environ.get("RSCIIO_TEST_DATA_SOURCE", "")
    if src:
        import shutil

        src = Path(src)
        dest = PATH / "data"
        print(f"Copying test data from {src} to {dest}")  # noqa: T201
        shutil.copytree(str(src), str(dest), dirs_exist_ok=True)
        n = len(list(dest.rglob("*")))
        print(f"Copied {n} entries to {dest}")  # noqa: T201

    from rsciio.tests.registry import TEST_DATA_REGISTRY

    if pooch_object is None:
        pooch_object = TEST_DATA_REGISTRY
    if ignore_hash is None:
        ignore_hash = sys.platform == "win32"
    if ignore_hash:
        for key in pooch_object.registry.keys():
            pooch_object.registry[key] = None

    if show_progressbar:
        try:
            from tqdm import tqdm

            pbar = tqdm(total=len(pooch_object.registry_files))
        except ImportError:
            print("Using progressbar requires the `tqdm` library.")  # noqa: T201
            show_progressbar = False

    for i, file in enumerate(pooch_object.registry_files):
        pooch_object.fetch(file, progressbar=False)
        if show_progressbar:
            pbar.update(i)

    if show_progressbar:
        pbar.close()


if __name__ == "__main__":
    # Used by the pre-commit hook
    update_registry()



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib
from pathlib import Path

import numpy as np
import pytest
from packaging.version import Version

imageio = pytest.importorskip("imageio")

from rsciio.image import file_writer  # noqa: E402

testfile_dir = (Path(__file__).parent / "data" / "image").resolve()


@pytest.mark.skipif(
    Version(imageio.__version__) < Version("2.23"),
    reason="needs imageio >=2.23",
)
@pytest.mark.parametrize(("dtype"), ["uint8", "int32", bool])
@pytest.mark.parametrize(("ext"), ["png", "bmp", "gif", "jpg"])
def test_save_load_cycle_grayscale(dtype, ext, tmp_path):
    if dtype == "int32" and ext in ["png", "bmp", "jpg"]:
        # PNG, BMP and JPG do not support int32.
        return

    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    s = hs.signals.Signal2D(np.arange(128 * 128).reshape(128, 128).astype(dtype))

    print(f"Saving-loading cycle for the extension `{ext}` with dtype `{dtype}`")  # noqa: T201
    filename = tmp_path / f"test_image.{ext}"
    s.save(filename)
    hs.load(filename)


@pytest.mark.parametrize(("color"), ["rgb8", "rgba8"])
@pytest.mark.parametrize(("ext"), ["png", "bmp", "gif", "jpeg"])
def test_save_load_cycle_color(color, ext, tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    dim = 4 if "rgba" in color else 3
    dtype = "uint8" if "8" in color else "uint16"
    if dim == 4 and ext == "jpeg":
        # JPEG does not support alpha channel.
        return

    print("color:", color, "; dim:", dim, "; dtype:", dtype)  # noqa: T201
    s = hs.signals.Signal1D(
        np.arange(128 * 128 * dim).reshape(128, 128, dim).astype(dtype)
    )
    s.change_dtype(color)

    print("Saving-loading cycle for the extension:", ext)  # noqa: T201
    filename = tmp_path / f"test_image.{ext}"
    s.save(filename)
    hs.load(filename)


@pytest.mark.skipif(
    Version(imageio.__version__) < Version("2.23"),
    reason="needs imageio >=2.23",
)
@pytest.mark.parametrize(("dtype"), ["uint8", "int32"])
@pytest.mark.parametrize(("ext"), ["png", "bmp", "gif", "jpg"])
def test_save_load_cycle_kwds(dtype, ext, tmp_path):
    if dtype == "int32" and ext in ["png", "bmp", "jpg"]:
        # PNG, BMP and JPG do not support int32.
        return

    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    s = hs.signals.Signal2D(np.arange(128 * 128).reshape(128, 128).astype(dtype))

    print(f"Saving-loading cycle for the extension `{ext}` with dtype `{dtype}`")  # noqa: T201
    filename = tmp_path / f"test_image.{ext}"
    if ext == "png":
        kwds = {"optimize": True}
    elif ext == "jpg":
        kwds = {"quality": 100, "optimize": True}
    elif ext == "gif":
        kwds = {"subrectangles": "True", "palettesize": 128}
    else:
        kwds = {}
    s.save(filename, **kwds)
    hs.load(filename, pilmode="L", as_grey=True)


@pytest.mark.skipif(
    Version(imageio.__version__) < Version("2.23"),
    reason="needs imageio >=2.23",
)
@pytest.mark.parametrize(("ext"), ["png", "bmp", "gif", "jpg"])
def test_export_scalebar(ext, tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pytest.importorskip("matplotlib_scalebar")
    import matplotlib.pyplot as plt

    # Use np.uint8 to be able to save as BMP
    pixels = 64
    data = np.arange(pixels**2).reshape((pixels, pixels)).astype(np.uint8)
    s = hs.signals.Signal2D(data)
    s.axes_manager[0].units = "nm"
    s.axes_manager[1].units = "nm"

    filename = tmp_path / f"test_scalebar_export.{ext}"
    if ext in plt.figure().canvas.get_supported_filetypes():
        s.save(filename, scalebar=True)
    else:
        # file format not supported by matplotlib backend
        with pytest.raises(ValueError):
            s.save(filename, scalebar=True)
        with pytest.raises(ValueError):
            s.save(filename, output_size=512)
        s.save(filename)

    s_reload = hs.load(filename)
    assert s.data.shape == s_reload.data.shape


@pytest.mark.parametrize(("units"), ["1/nm", "1 / nm", "1 / nanometer", "1/nanometer"])
def test_export_scalebar_reciprocal(tmp_path, units):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = 512
    s = hs.signals.Signal2D(
        np.arange(pixels**2).reshape((pixels, pixels)).astype("int32")
    )
    for axis in s.axes_manager.signal_axes:
        axis.units = units
        axis.scale = 0.1

    filename = tmp_path / "test_scalebar_export.png"
    s.save(filename, scalebar=True, scalebar_kwds={"location": "lower right"})
    s_reload = hs.load(filename)
    assert s.data.shape == s_reload.data.shape


def test_export_scalebar_undefined_units(tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = 512
    s = hs.signals.Signal2D(
        np.arange(pixels**2).reshape((pixels, pixels)).astype("int32")
    )

    filename = tmp_path / "test_scalebar_export.png"
    s.save(filename, scalebar=True, scalebar_kwds={"location": "lower right"})
    s_reload = hs.load(filename)
    assert s.data.shape == s_reload.data.shape


def test_non_uniform(tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = 16
    s = hs.signals.Signal2D(np.arange(pixels**2).reshape((pixels, pixels)))
    s.axes_manager[0].convert_to_non_uniform_axis()

    filename = tmp_path / "test_export_size.jpg"
    with pytest.raises(TypeError):
        s.save(filename)


def test_export_scalebar_different_scale_units(tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pytest.importorskip("matplotlib_scalebar")
    pixels = 16
    s = hs.signals.Signal2D(np.arange(pixels**2).reshape((pixels, pixels)))
    s.axes_manager[0].scale = 2

    filename = tmp_path / "test_export_scalebar_different_scale_units0.jpg"
    with pytest.raises(ValueError):
        s.save(filename, scalebar=True)

    s = hs.signals.Signal2D(np.arange(pixels**2).reshape((pixels, pixels)))
    s.axes_manager[0].units = "nm"

    filename = tmp_path / "test_export_scalebar_different_scale_units1.jpg"
    with pytest.raises(ValueError):
        s.save(filename, scalebar=True)


@pytest.mark.parametrize(
    "pixels_below_above", [(10, 0.991, 0.989), (1000, 0.99991, 0.99989)]
)
def test_export_scalebar_scale_close_enough(tmp_path, pixels_below_above):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pytest.importorskip("matplotlib_scalebar")
    pixels, below, above = pixels_below_above
    s = hs.signals.Signal2D(np.arange(pixels**2).reshape((pixels, pixels)))
    # scale is not exactly equal, but close enough
    s.axes_manager[0].scale = below
    s.axes_manager[1].scale = 1.0
    s.axes_manager.signal_axes.set(units="nm")

    filename = tmp_path / "test_export_scalebar_scale_close_enough0.jpg"
    s.save(filename, scalebar=True)
    _ = hs.load(filename)

    # scale is not close enough above the limit of the relative error
    s.axes_manager[0].scale = above
    filename = tmp_path / "test_export_scalebar_scale_close_enough1.jpg"
    with pytest.raises(ValueError):
        s.save(filename, scalebar=True)


def test_save_image_scalebar_scale_opposite_signs(tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = 16
    s = hs.signals.Signal2D(
        np.arange(pixels**2).reshape((pixels, pixels)).astype("int32")
    )
    s.axes_manager[0].scale = -1.0
    s.axes_manager[1].scale = 1.0

    fname = tmp_path / "test_save_image_scalebar_scale_opposite_signs.png"
    s.save(fname, scalebar=True)


@pytest.mark.parametrize("output_size", (512, [512, 512]))
def test_export_output_size(output_size, tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = 16
    s = hs.signals.Signal2D(np.arange(pixels**2).reshape((pixels, pixels)))

    fname = tmp_path / "test_export_size.jpg"
    s.save(fname, scalebar=True, output_size=output_size)
    s_reload = hs.load(fname)
    assert s_reload.data.shape == (512, 512)


@pytest.mark.parametrize("scalebar", [True, False])
@pytest.mark.parametrize("output_size", (None, 512, (512, 512)))
def test_export_output_size_non_square(output_size, tmp_path, scalebar):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = (8, 16)
    s = hs.signals.Signal2D(
        np.arange(np.multiply(*pixels), dtype=np.uint8).reshape(pixels)
    )

    fname = tmp_path / "test_export_size_non_square.jpg"
    s.save(fname, output_size=output_size, scalebar=scalebar)
    s_reload = hs.load(fname)

    if output_size is None:
        output_size = (8, 16)
    if isinstance(output_size, int):
        output_size = (output_size * np.divide(*pixels), output_size)

    assert s_reload.data.shape == output_size


@pytest.mark.parametrize("output_size", (None, 512))
@pytest.mark.parametrize("aspect", (1, 0.5))
def test_export_output_size_aspect(aspect, output_size, tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = (256, 256)
    s = hs.signals.Signal2D(np.arange(np.multiply(*pixels)).reshape(pixels))

    fname = tmp_path / "test_export_size_non_square_aspect.jpg"
    s.save(
        fname,
        scalebar=True,
        output_size=output_size,
        imshow_kwds={"aspect": aspect},
        scalebar_kwds={"rotation": "horizontal-only"},
    )
    s_reload = hs.load(fname)

    if output_size is None:
        output_size = s.data.shape[0]
    assert s_reload.data.shape == (output_size * aspect, output_size)


def test_save_image_navigation(tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = 16
    s = hs.signals.Signal2D(
        np.arange(pixels**2).reshape((pixels, pixels)).astype("int32")
    )

    fname = tmp_path / "test_save_image_navigation.png"
    s.T.save(fname, scalebar=True)


def test_error_library_no_installed(tmp_path):
    axis = {
        "_type": "UniformDataAxis",
        "name": None,
        "units": None,
        "navigate": False,
        "is_binned": False,
        "size": 128,
        "scale": 1.0,
        "offset": 0.0,
    }
    signal_dict = {"data": np.arange(128 * 128).reshape(128, 128), "axes": [axis, axis]}

    matplotlib = importlib.util.find_spec("matplotlib")
    if matplotlib is None:
        # When matplotlib is not installed, raises an error to inform user
        # that matplotlib is necessary
        with pytest.raises(ValueError):
            file_writer(tmp_path / "test_image_error.jpg", signal_dict, output_size=64)

        with pytest.raises(ValueError):
            file_writer(
                tmp_path / "test_image_error.jpg", signal_dict, imshow_kwds={"a": "b"}
            )


def test_renishaw_wire():
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    s = hs.load(testfile_dir / "renishaw_wire.jpg")
    assert s.data.shape == (480, 752)
    for axis, scale, offset, name in zip(
        s.axes_manager.signal_axes,
        [2.42207446, 2.503827],
        [19105.5, -6814.538],
        ["y", "x"],
    ):
        np.testing.assert_allclose(axis.scale, scale)
        np.testing.assert_allclose(axis.offset, offset)
        axis.name == name
        axis.units == "µm"


def test_export_output_size_iterable_length_1(tmp_path):
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    pixels = (256, 256)
    s = hs.signals.Signal2D(np.arange(np.multiply(*pixels)).reshape(pixels))

    fname = tmp_path / "test_export_output_size_iterable_length_1.jpg"
    with pytest.raises(ValueError):
        s.save(fname, output_size=(256,))


def test_missing_exif_tags():
    hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
    import traits.api as t

    s = hs.load(testfile_dir / "jpg_no_exif_tags.jpg")

    assert s.data.shape == (182, 255)
    assert s.axes_manager.signal_shape == (255, 182)
    for axis in s.axes_manager.signal_axes:
        assert axis.scale == 1
        assert axis.offset == 0
        assert axis.units == t.Undefined
